@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2013 Christoph Ender.
+ * Copyright (c) 2009-2014 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -69,7 +69,7 @@
 
 #include "../locales/fizmo_ncursesw_locales.h"
 
-#define FIZMO_NCURSESW_VERSION "0.7.7"
+#define FIZMO_NCURSESW_VERSION "0.7.8"
 
 #ifdef ENABLE_X11_IMAGES
 #include <drilbo/drilbo.h>
@@ -1723,11 +1723,11 @@ void copy_area(int dsty, int dstx, int srcy, int srcx, int height, int width)
   int y, x;
   cchar_t wcval;
 
-  TRACE_LOG("copyarea:%d/%d/%d/%d/%d/%d\n", 
-      srcy,
-      srcx,
+  TRACE_LOG("copyarea, dst:%d/%d, src:%d/%d, size:%d*%d\n", 
       dsty,
       dstx,
+      srcy,
+      srcx,
       height,
       width);
 
