@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2013 Christoph Ender.
+ * Copyright (c) 2009-2014 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -892,7 +892,7 @@ int save_game_to_stream(uint16_t address, uint16_t length, z_file *save_file,
 
       do
       {
-        return_code = output_rewind_paragraph(history);
+        return_code = output_rewind_paragraph(history, NULL);
         nof_paragraphs_to_save--;
       }
       while ( (nof_paragraphs_to_save > 0) && (return_code == 0) );
@@ -1588,6 +1588,10 @@ int restore_game_from_stream(uint16_t address, uint16_t length,
 
   free(restored_story_mem);
 
+  fizmo_new_screen_size(
+      active_interface->get_screen_width_in_characters(),
+      active_interface->get_screen_height_in_lines());
+
 #ifdef ENABLE_TRACING
   dump_stack_to_tracelog();
   dump_dynamic_memory_to_tracelog();
