@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2013 Christoph Ender.
+ * Copyright (c) 2009-2014 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -194,7 +194,7 @@ static int zchar_storage_write(unsigned int five_bits)
 
   if (++zchar_storage_word_index == 3)
   {
-    TRACE_LOG("zchar: storing word %x at %x, end at %x.\n",
+    TRACE_LOG("zchar: storing word %x at %p, end at %p.\n",
         zchar_storage_word,
         zchar_storage_output_index,
         zchar_storage_index_behind);
@@ -268,7 +268,7 @@ static int zchar_storage_finish()
     {
       // In case there's some output, write the string termination to the
       // last word of the output.
-      TRACE_LOG("Successfully finalized full word.");
+      TRACE_LOG("Successfully finalized full word.\n");
       *(zchar_storage_output_index-2) |= 0x80;
     }
     else
@@ -619,8 +619,8 @@ static uint8_t locate_dictionary_entry(
     if (dont_write_unrecognized_words_to_parse_buffer == false)
     {
       store_word(*parse_position, 0);
-      (*parse_position) += 2;
     }
+    (*parse_position) += 2;
   }
 
   if (dont_write_unrecognized_words_to_parse_buffer == false)
@@ -636,6 +636,10 @@ static uint8_t locate_dictionary_entry(
     **parse_position = word_position;
     (*parse_position)++;
   }
+  else
+  {
+    (*parse_position) += 2;
+  }
 
   if (word_found_at_index != -1)
   {
@@ -693,6 +697,9 @@ static void tokenise(
   dictionary += number_of_input_codes;
   dictionary_entry_length = *(dictionary++);
   number_of_dictionary_entries = (int16_t)load_word(dictionary);
+  TRACE_LOG("Number of dictionary entries: %d.\n",
+      number_of_dictionary_entries);
+
   if (number_of_dictionary_entries < 0)
   {
     dictionary_is_unsorted = true;
@@ -1511,6 +1518,14 @@ static bool process_interpreter_command()
   {
     active_interface->output_interface_info();
 
+#ifdef FIZMO_DIST_VERSION
+    i18n_translate(
+        libfizmo_module_name,
+        i18n_libfizmo_FIZMO_VERSION_P0S,
+        FIZMO_DIST_VERSION);
+    streams_latin1_output("\n");
+#endif // FIZMO_DIST_VERSION
+
     (void)i18n_translate(
         libfizmo_module_name,
         i18n_libfizmo_LIBFIZMO_VERSION_P0S,
@@ -1814,6 +1829,7 @@ int read_command_from_file(zscii *input_buffer, int input_buffer_size,
   int milliseconds;
   int res;
   int return_code;
+  z_ucs unicode_input;
 
   if (input_stream_1 == NULL)
   {
@@ -1867,34 +1883,29 @@ int read_command_from_file(zscii *input_buffer, int input_buffer_size,
     fsi->setfilepos(input_stream_1, filepos, SEEK_SET);
   }
 
-  // FIMXE: Input conversion from latin1(?) to zscii-input.
-  while (
-      ((c = fsi->readchar(input_stream_1)) != EOF)
-      &&
-      (c != '\n')
-      &&
-      (input_length != input_buffer_size)
-      )
-  {
-    if (c != '\r')
-    {
-      *(input_buffer++) = unicode_char_to_zscii_input_char(c & 0x7f);
+  while (input_length != input_buffer_size) {
+    unicode_input = parse_utf8_char_from_file(input_stream_1);
+
+    if (unicode_input == '\n')
+      break;
+    else if (unicode_input != '\r') {
+      *(input_buffer++) = unicode_char_to_zscii_input_char(unicode_input);
       input_length++;
     }
   }
 
-  if (c != EOF)
+  if (unicode_input != UEOF)
   {
     // Seek next '\n'.
-    while (c != '\n')
-      if ((c = fsi->readchar(input_stream_1)) == EOF)
+    while (unicode_input != '\n')
+      if ((unicode_input = parse_utf8_char_from_file(input_stream_1)) == UEOF)
         break;
   }
 
   TRACE_LOG("Read %d input chars.\n", input_length);
 
   // At EOF? If so, close stream and return last length.
-  if (c == EOF)
+  if (unicode_input == UEOF)
   {
     fsi->closefile(input_stream_1);
     input_stream_1_active = false;
@@ -2000,14 +2011,6 @@ void opcode_read(void)
       // verification_routine are set to a fixed 0 in the call above)
       // the result stored in input_length is always >= 0.
 
-      // Reduce to lower case
-      // FIXME: Respect other (non-ASCII) languages?
-      /*
-      for (i=1; i<input_length+1; i++)
-        if ((z_text_buffer[i] >= 0x41) && (z_text_buffer[i] <= 0x5a))
-          z_text_buffer[i] += 0x20;
-      */
-
       z_text_buffer[input_length + 1] = 0;
 
 #ifndef DISABLE_COMMAND_HISTORY
@@ -2218,8 +2221,18 @@ void opcode_read(void)
         stream_4_latin1_output("(Interrupted input)");
     }
 
-    if (input_length >= 0)
+    if (input_length >= 0) {
       (void)streams_z_ucs_output_user_input(interpreter_command_buffer);
+      // Reduce to lower case. This appears to be also required in v5
+      // games, although it's not explicitely stated in the Z-Spec (see
+      // Beyond Zork, character name: "frank booth" becomes "Frank Booth").
+      // ToDo: Respect other (non-ASCII) languages?
+      for (i=0; i<input_length; i++)
+        if ((z_text_buffer[offset + i] >= 0x41)
+            && (z_text_buffer[offset + i] <= 0x5a))
+          z_text_buffer[offset + i] += 0x20;
+    }
+
     (void)streams_z_ucs_output_user_input(z_ucs_newline_string);
 
     stream_1_active = stream_1_active_buf;
@@ -2540,18 +2553,29 @@ void opcode_print_table(void)
 void opcode_encode_text(void)
 {
   uint8_t *src = z_mem + (uint16_t)op[0] + (uint16_t)op[2];
-  uint16_t length = (uint16_t)op[1];
+  uint16_t length = (uint16_t)op[1], chars_written = 0;
   uint8_t *dest = z_mem + (uint16_t)op[3];
 
   TRACE_LOG("Opcode: ENCODE_TEXT.\n");
 
-  zchar_storage_start(dest, (length + 2) / 3);
+  // "encode_text" is only valid for version >= 5, so the length of the
+  // buffer must always be fixed at 6 bytes.
+  zchar_storage_start(dest, 6);
+  TRACE_LOG("Encoding to %x / %p.\n", (uint16_t)op[3], dest);
 
-  while ( (*src != 0) && (length > 0) )
+  while ( (*src != 0) && (chars_written < length) )
   {
+    TRACE_LOG("Encoding char '%c'.\n", *src);
     store_ZSCII_as_zchar(*src);
     src++;
-    length--;
+    chars_written++;
+  }
+  while (chars_written < 9) {
+    // Since we're always working in version >= 5, we always have to pad
+    // to a fixed size of 9 z-chars.
+    TRACE_LOG("Storing padding char 5.\n");
+    (void)zchar_storage_write(5);
+    chars_written++;
   }
 
   zchar_storage_finish();
