@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2013 Christoph Ender.
+ * Copyright (c) 2009-2014 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -69,6 +69,8 @@
 //static z_ucs word_interpunctation_chars[] = {
 //  Z_UCS_MINUS, Z_UCS_DOT, 0 };
 
+static z_ucs word_split_chars[] = { Z_UCS_SPACE, Z_UCS_NEWLINE, 0 };
+
 
 WORDWRAP *wordwrap_new_wrapper(size_t line_length,
     void (*wrapped_text_output_destination)(z_ucs *output, void *parameter),
@@ -200,7 +202,7 @@ static void output_buffer(WORDWRAP *wrapper, z_ucs *buffer_start,
 static void flush_input_buffer(WORDWRAP *wrapper, bool force_flush)
 {
   z_ucs *index = NULL, *ptr, *hyphenated_word, *last_hyphen, *word_start;
-  z_ucs *word_end, *input = wrapper->input_buffer, *first_space;
+  z_ucs *word_end, *input = wrapper->input_buffer, *first_space_or_newline;
   z_ucs buf=0, buf2; // buf initialized to avoid compiler warning
   z_ucs buf3 = '-'; // buf3 initialized to avoid compiler warning
   long len, chars_sent = 0;
@@ -216,7 +218,8 @@ static void flush_input_buffer(WORDWRAP *wrapper, bool force_flush)
 
   for (;;)
   {
-    TRACE_LOG("Processing flush for line-length %d\n", current_line_length);
+    TRACE_LOG("Processing flush for line-length %d, already in line: %d.\n",
+        current_line_length, wrapper->chars_already_on_line);
 
     if (*input != 0)
     {
@@ -328,8 +331,8 @@ static void flush_input_buffer(WORDWRAP *wrapper, bool force_flush)
         last_hyphen = NULL;
       }
       else {
-        if ( ((first_space = z_ucs_chr(
-                  input + current_line_length, Z_UCS_SPACE)) == NULL) 
+        if ( ((first_space_or_newline = z_ucs_chrs(
+                  input + current_line_length, word_split_chars)) == NULL) 
             && (force_flush == false) ) {
           // In case we can't find any space the word may not have been
           // completely written to the buffer. Wait until we've got more
@@ -338,7 +341,7 @@ static void flush_input_buffer(WORDWRAP *wrapper, bool force_flush)
           break;
         }
         else {
-          if (first_space == NULL) {
+          if (first_space_or_newline == NULL) {
             word_end = input + current_line_length;
             while (*(word_end + 1) != 0) {
               word_end++;
@@ -347,7 +350,7 @@ static void flush_input_buffer(WORDWRAP *wrapper, bool force_flush)
           else {
             // We've found a space behind the overrunning word so we're
             // able to correctly split the current line.
-            word_end = first_space - 1;
+            word_end = first_space_or_newline - 1;
           }
 
           // Before hyphentation, check for dashes inside the last word.
@@ -426,9 +429,25 @@ static void flush_input_buffer(WORDWRAP *wrapper, bool force_flush)
               index = last_hyphen + 1;
             }
             else {
+              // We couldn't find a possibility to hyphenate the last
+              // word in the line.
               TRACE_LOG("No hyphen found.\n");
-              if ( (index > input) && (*(index-1) != Z_UCS_MINUS) ) {
-                index--;
+
+              if (index > input) {
+                if  (*(index-1) != Z_UCS_MINUS) {
+                  // In case the char before the last word is not a dash,
+                  // we'll skip the space before this word by moving back
+                  // the index by one.
+                  index--;
+                }
+              }
+              else {
+                // In case the current word is so long that it doesn't fit
+                // on the line -- this may be the case if we're supposed
+                // to display ASCII art and the linesize is to short -- we
+                // have to advance the index to the line end.
+                TRACE_LOG("This is the first word in the line, hard break.\n");
+                index = input + current_line_length;
               }
             }
           }
@@ -464,7 +483,8 @@ static void flush_input_buffer(WORDWRAP *wrapper, bool force_flush)
         index--;
       }
 
-      if (*index == Z_UCS_SPACE) {
+      // if (*index == Z_UCS_SPACE) {
+      while (*index == Z_UCS_SPACE) {
         index++;
       }
 
@@ -515,7 +535,8 @@ static void flush_input_buffer(WORDWRAP *wrapper, bool force_flush)
       *(index+1) = buf2;
       *index = buf;
 
-      if (*index == Z_UCS_SPACE) {
+      //if (*index == Z_UCS_SPACE) {
+      while (*index == Z_UCS_SPACE) {
         index++;
       }
 
@@ -693,5 +714,10 @@ void wordwrap_remove_chars(WORDWRAP *wrapper, size_t num_chars_to_remove)
   }
 }
 
+void wordwrap_set_line_index(WORDWRAP *wrapper, int new_line_index) {
+  TRACE_LOG("Setting chars on line to %d.\n", new_line_index);
+  wrapper->chars_already_on_line = new_line_index;
+}
+
 #endif /* wordwrap_c_INCLUDED */
 
