@@ -3,7 +3,7 @@
  *
  * This file is part of fizmo.
  *
- * Copyright (c) 2009-2013 Christoph Ender.
+ * Copyright (c) 2009-2014 Christoph Ender.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -91,7 +91,7 @@ void wordwrap_flush_output(WORDWRAP *wrapper);
 void wordwrap_insert_metadata(WORDWRAP *wrapper,
     void (*metadata_output)(void *ptr_parameter, uint32_t int_parameter),
     void *ptr_parameter, uint32_t int_parameter);
-//void wordwrap_set_line_index(WORDWRAP *wrapper, int new_line_index);
+void wordwrap_set_line_index(WORDWRAP *wrapper, int new_line_index);
 //void wordwrap_output_left_padding(WORDWRAP *wrapper);
 void wordwrap_adjust_line_length(WORDWRAP *wrapper, size_t new_line_length);
 // To remove chars from the end of the current(!) line:
